# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

from .Model import Model, SaveCallback, ProgressCallback, \
        ReloadDatasetCallback, RandomDatasetCallback, \
        ReAccuracyCallback
from .utils import compile, fit
from .csv import CSVReader

__version__ = "0.1.13"

__all__ = (
    "Model",
    "SaveCallback",
    "ProgressCallback",
    "ReAccuracyCallback",
    "RandomDatasetCallback",
    "ReloadDatasetCallback",
    "CSVReader",
    "compile",
    "fit",
)



# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch, time, json, math, gc, os
import numpy as np
from torch.utils.data import DataLoader, Dataset
from .utils import TransformDataset, list_files, \
    get_default_device, batch_to, tensor_size, \
    load_json, summary, fit


class Model:
    
    def __init__(self, module=None):
        self.device = 'cpu'
        self.transform_x = None
        self.transform_y = None
        self.module = module
        self.optimizer = None
        self.scheduler = None
        self.loss = None
        self.loss_reduction = 'mean'
        self.best_metrics = ["epoch"]
        self.acc_fn = None
        self.acc_reduction = 'sum'
        self.name = module.__class__.__name__
        self.prefix_name = ""
        self.epoch = 1
        self.history = {}
        self.min_lr = 1e-5
        self.model_path = ""
        self.repository_path = ""
        self.set_repository_path("model")
    
    
    def set_module(self, module):
        self.module = module
        return self
    
    def set_optimizer(self, optimizer):
        self.optimizer = optimizer
        return self
    
    def set_loss(self, loss, reduction = 'mean'):
        self.loss = loss
        self.loss_reduction = reduction
        return self
    
    def set_scheduler(self, scheduler):
        self.scheduler = scheduler
        return self
    
    def set_acc(self, acc, reduction = 'sum'):
        self.acc_fn = acc
        self.acc_reduction = reduction
        return self
    
    def set_name(self, name):
        self.name = name
        self.set_repository_path(self.repository_path)
        return self
    
    def set_prefix_name(self, prefix_name):
        self.prefix_name = prefix_name
        self.set_repository_path(self.repository_path)
        return self
    
    def set_path(self, model_path):
        self.model_path = model_path
        return self
    
    def set_best_metrics(self, best_metrics):
        self.best_metrics = best_metrics
        return self
    
    def set_repository_path(self, repository_path):
        self.repository_path = repository_path
        self.model_path = os.path.join(repository_path, self.get_model_name())
        return self
    
    def get_model_name(self):
        if self.prefix_name != "":
            return self.name + "_" + self.prefix_name
        return self.name
    
    def get_epoch(self):
        return self.epoch - 1
    
    
    def to(self, device):
        self.module = self.module.to(device)
        self.device = device
        return self
    
    def to_cuda(self):
        self.to( torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu') )
        return self
    
    def to_cpu(self):
        self.to( torch.device("cpu") )
        return self
    
    def train(self):
        self.module.train()
        return self
    
    def eval(self):
        self.module.eval()
        return self
    
    
    def load_state_dict(self, save_metrics, strict=False):
        
        if "epoch" in save_metrics:
            self.epoch = save_metrics["epoch"] + 1
            
            # Load history
            if "history" in save_metrics:
                self.history = save_metrics["history"].copy()
            
            # Load module
            if "module" in save_metrics:
                state_dict = save_metrics["module"]
                self.module.load_state_dict(state_dict, strict=strict)
            
            # Load optimizer
            if "optimizer" in save_metrics:
                state_dict = save_metrics["optimizer"]
                self.optimizer.load_state_dict(state_dict)
            
            # Load scheduler
            if "scheduler" in save_metrics:
                state_dict = save_metrics["scheduler"]
                self.scheduler.load_state_dict(state_dict)
            
            # Load loss
            #if "loss" in save_metrics:
            #    state_dict = save_metrics["loss"]
            #    self.loss.load_state_dict(state_dict)
        
        else:
            self.module.load_state_dict(save_metrics, strict=strict)
        
        return self
    
    
    def load_model(self, file_path, full_path=False):
        
        """
        Load model from file
        """
        
        if not os.path.exists(file_path) and not full_path:
            file_path = os.path.join(self.model_path, file_path)
        
        save_metrics = torch.load(file_path)
        self.load_state_dict(save_metrics)
        
        return self
        
    
    def load_epoch(self, epoch):
        
        """
        Load epoch
        """
        
        file_name = self.get_model_name() + "-" + str(epoch) + ".data"
        file_path = os.path.join(self.model_path, file_name)
        
        if not os.path.exists(file_path):
            file_name = self.get_model_name() + "-" + str(epoch) + ".pth"
            file_path = os.path.join(self.model_path, file_name)
        
        self.load_model(file_path, full_path=True)
        
        return self
        
        
    def load_last(self):
        
        """
        Load last model
        """
        
        file_name = self.get_model_name() + ".data"
        file_path = os.path.join(self.model_path, file_name)
        
        if not os.path.exists(file_path):
            file_name = self.get_model_name() + ".pth"
            file_path = os.path.join(self.model_path, file_name)
        
        if os.path.exists(file_path):
            self.load_model(file_path, full_path=True)
            return self
        
        file_name = os.path.join(self.model_path, "history.json")
        if os.path.exists(file_name):
        
            obj = load_json(file_name)
            
            if obj is not None:
                epoch = obj["epoch"]
                self.load_epoch(epoch)
        
        return self
        
    
    def load_best(self):
        
        """
        Load best model
        """
        
        file_path = os.path.join(self.model_path, "history.json")
        
        if not os.path.exists(file_path):
            return self
        
        obj = load_json(file_path)
        
        if obj is not None:
            best_epoch = obj["best_epoch"]
            self.load_epoch(best_epoch)
        
        return self
    
    
    def save_weights_epoch(self, file_path=None):
        
        """
        Save weights
        """
        
        # Create folder
        if not os.path.isdir(self.model_path):
            os.makedirs(self.model_path)
        
        if file_path is None:
            file_name = self.get_model_name() + "-" + str(self.epoch) + ".pth"
            file_path = os.path.join(self.model_path, file_name)
        
        torch.save(self.module.state_dict(), file_path)
        
        return self
    
    
    def save_weights(self, file_path=None):
        
        """
        Save weights
        """
        
        # Create folder
        if not os.path.isdir(self.model_path):
            os.makedirs(self.model_path)
        
        if file_path is None:
            file_name = self.get_model_name() + ".pth"
            file_path = os.path.join(self.model_path, file_name)
        
        torch.save(self.module.state_dict(), file_path)
        
        return self
    
    
    def save_train_epoch(self):
        
        """
        Save train epoch
        """
        
        # Create folder
        if not os.path.isdir(self.model_path):
            os.makedirs(self.model_path)
        
        file_name = self.get_model_name() + "-" + str(self.epoch) + ".data"
        file_path = os.path.join(self.model_path, file_name)
        self.save_model(file_path)
        
        return self
    
        
    def save_model(self, file_path=None):
        
        """
        Save train status
        """
        
        # Get metrics
        save_metrics = {}
        save_metrics["name"] = self.get_model_name()
        save_metrics["epoch"] = self.epoch
        save_metrics["history"] = self.history.copy()
        save_metrics["module"] = self.module.state_dict()
        
        if self.optimizer is not None:
            save_metrics["optimizer"] = self.optimizer.state_dict()
        
        if self.scheduler is not None:
            save_metrics["scheduler"] = self.scheduler.state_dict()
        
        #if self.loss is not None:
        #    save_metrics["loss"] = self.loss.state_dict()
        
        # Create folder
        if not os.path.isdir(self.model_path):
            os.makedirs(self.model_path)
        
        # Save model to file
        if file_path is None:
            model_file_name = self.get_model_name() + ".data"
            file_path = os.path.join(self.model_path, model_file_name)
        
        torch.save(save_metrics, file_path)
        
        return self
    
    
    def save_history(self):
        
        """
        Save history to json
        """
        
        best_epoch = self.get_the_best_epoch(self.best_metrics)
        file_name = os.path.join(self.model_path, "history.json")
        obj = {
            "epoch": self.epoch,
            "best_epoch": best_epoch,
            "history": self.history.copy(),
        }
        json_str = json.dumps(obj, indent=2)
        file = open(file_name, "w")
        file.write(json_str)
        file.close()
        
        return self
    
    
    def do_training(self, max_epochs):
        
        """
        Returns True if model is need to train
        """
        
        if self.epoch > max_epochs:
            return False
        
        for item in self.optimizer.param_groups:
            if item["lr"] >= self.min_lr:
                return True
        
        return False
    
    
    def set_new_lr(self, lr):
        
        for index, param_group in enumerate(self.optimizer.param_groups):
            param_group['lr'] = lr[index]
        
        return self
    
    
    def __call__(self, x):
        return self.module(x)
    
    
    def predict(self, x):
        
        """
        Predict
        """
        
        batch = {"x":x}
        batch_transform = getattr(self.module, "batch_transform", None)
        
        with torch.no_grad():
            
            self.module.eval()
            
            if batch_transform:
                batch = batch_transform(batch, self.device)
            
            x = batch["x"]
            
            if hasattr(self.module, "step_forward"):
                _, _, y_predict = self.module.step_forward(x, self.device)
            
            else:
                x = batch_to(x, self.device)
                y = self.module(x)
        
        return y
    
    
    def predict_dataset(self, dataset, predict, batch_size=64, obj=None, collate_fn=None):
        
        """
        Predict dataset
        """
        
        loader = None
        device = self.device
        
        if isinstance(dataset, Dataset):
            loader = DataLoader(
                dataset,
                batch_size=batch_size,
                collate_fn=collate_fn,
                drop_last=False,
                shuffle=False
            )
        
        if isinstance(dataset, DataLoader):
            loader = dataset
        
        batch_transform = getattr(self.module, "batch_transform", None)
        
        with torch.no_grad():
        
            self.module.eval()
            
            pos = 0
            next_pos = 0
            dataset_count = len(dataset)
            time_start = time.time()
            
            for batch in loader:
                
                if batch_transform:
                    batch = batch_transform(batch, self.device)
                
                if hasattr(self.module, "step_forward"):
                    _, _, y_predict = self.module.step_forward(batch, self.device)
                
                else:
                    x_batch = batch_to(batch["x"], device)
                    y_predict = self.module(x_batch)
                    del x_batch
                
                predict(batch, y_predict, obj)
                
                # Show progress
                pos = pos + len(batch['x'])
                if pos > next_pos:
                    next_pos = pos + 16
                    t = str(round(time.time() - time_start))
                    print (
                        "\r" + str(math.floor(pos / dataset_count * 10000) / 100) + "% " +
                        t + "s", end=''
                    )
                
                del y_predict, batch
                
                # Clear cache
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
                
                gc.collect()
            
            print ("\nOk")
    
    
    def get_metrics(self, metric_name, convert=False):
        
        """
        Returns metrics by name
        """
        
        def convert_value(value, metric_name):
            if (
                metric_name == "train_acc" or
                metric_name == "train_acc_value" or
                metric_name == "val_acc" or
                metric_name == "val_acc_value" or
                metric_name == "epoch"
            ):
                return -value
            return value
        
        res = []
        epochs = list(self.history.keys())
        for index in epochs:
            
            epoch = self.history[index]
            res2 = [ index ]
            
            if isinstance(metric_name, list):
                for name in metric_name:
                    value = epoch[name] if name in epoch else 0
                    if convert:
                        value = convert_value(value, name)
                    res2.append( value )
            
            else:
                value = epoch[metric_name] if metric_name in epoch else 0
                if convert:
                    value = convert_value(value, metric_name)
                res2.append( value )
            
            res.append(res2)
            
        return res
    
    
    def get_metric(self, metric_name, convert=False):
        
        """
        Returns metrics by name
        """
        
        res = []
        epochs = list(self.history.keys())
        for index in epochs:
            
            epoch = self.history[index]
                        
            value = epoch[metric_name] if metric_name in epoch else 0
            if convert:
                value = convert_value(value, metric_name)
            res.append( value )
            
        return res
    
    
    def get_the_best_epoch(self, best_metrics=None):
        
        """
        Returns the best epoch
        """
        
        epoch_indexes = self.get_the_best_epochs_indexes(1, best_metrics)
        best_epoch = epoch_indexes[0] if len(epoch_indexes) > 0 else 0
        return best_epoch
    
    
    def get_the_best_epochs_indexes(self, epoch_count=5, best_metrics=None):
        
        """
        Returns best epoch indexes
        """
        
        if best_metrics is None:
            best_metrics = self.best_metrics
        
        metrics = self.get_metrics(best_metrics, convert=True)
        metrics.sort(key=lambda x: x[1:])
        
        res = []
        res_count = 0
        metrics_len = len(metrics)
        loss_val_last = 100
        for index in range(metrics_len):
            
            res.append( metrics[index] )
            
            if loss_val_last != metrics[index][1]:
                res_count = res_count + 1
            
            loss_val_last = metrics[index][1]
            
            if res_count > epoch_count:
                break
        
        res = [ res[index][0] for index in range(len(res)) ]
        
        return res
    
    
    def get_best_epoch(self, best_metrics=None):
        
        """
        Returns the best epoch
        """
        
        return self.get_the_best_epoch(best_metrics)
    
    
    def get_best_epochs(self, epoch_count=5, best_metrics=None):
        
        """
        Returns best epoch indexes
        """
        
        return self.get_the_best_epochs_indexes(epoch_count, best_metrics)
    
    
    def save_the_best_models(self, max_best_models=10):
        
        """
        Save the best models
        """
        
        def detect_type(file_name):
            
            import re
            
            file_type = ""
            epoch_index = 0
            model_name = self.get_model_name()
            
            result = re.match(r'^'+model_name+'-(?P<id>[0-9]+)\.data$', file_name)
            if result:
                return "model", int(result.group("id"))
            
            result = re.match(r'^'+model_name+'-(?P<id>[0-9]+)\.pth$', file_name)
            if result:
                return "model", int(result.group("id"))
            
            return file_type, epoch_index
        
        
        if self.epoch > 0 and max_best_models > 0 and os.path.isdir(self.model_path):
            
            epoch_indexes = self.get_the_best_epochs_indexes(max_best_models)
            epoch_indexes.append( self.epoch )
            
            files = list_files( self.model_path )
            
            for file_name in files:
                
                file_type, epoch_index = detect_type(file_name)
                if file_type in ["model"] and \
                    epoch_index > 0 and \
                    not (epoch_index in epoch_indexes):
                    
                    file_path = os.path.join( self.model_path, file_name )
                    os.unlink(file_path)
    
    
    def summary(self, x, batch_size=2, collate_fn=None):
        
        """
        Show model summary
        """
        
        summary(self.module, x,
            device=self.device,
            collate_fn=collate_fn,
            model_name=self.get_model_name(),
            batch_size=batch_size
        )
    
        
    def draw_history_ax(self, ax, metrics=[], label=None, legend=True, convert=None, start=0):
        
        """
        Draw history to axes
        """
        
        metrics_values = self.get_metrics(metrics)
        metrics_values = metrics_values[start:]
        for index, name in enumerate(metrics):
            values = [ item[index + 1] for item in metrics_values ]
            if convert:
                values = list(map(convert, values))
            if len(values) > 0 and values[0] is not None:
                ax.plot( values, label=name)
        
        if label:
            ax.set_xlabel( label )
        
        if legend:
            ax.legend()
    
    
    def draw_history(self, show_acc=False, show_loss=True, start=0):
        
        """
        Draw history
        """
        
        import matplotlib.pyplot as plt
        
        pos = 0
        fig, ax = plt.subplots(1, show_acc + show_loss, figsize=(10, 4))
        if show_acc:
            self.draw_history_ax(
                ax[pos] if isinstance(ax, np.ndarray) else ax,
                ["train_acc", "val_acc"],
                label="Accuracy",
                convert=lambda x: x * 100 if x is not None else None,
                start=start
            )
            pos += 1
        if show_loss:
            self.draw_history_ax(
                ax[pos] if isinstance(ax, np.ndarray) else ax,
                ["train_loss", "val_loss"],
                label="Loss",
                start=start
            )
        plt.show()
    
    
    def print_history(self, progress):
        
        h = list(self.history.keys())
        h.sort()
        
        for epoch in h:
            s = progress.get_epoch_string(self.history[epoch].copy())
            print(s)
    
    
    def get_epoch_train_status(self):
        
        status = {
            "epoch": self.epoch,
            "time_start": 0,
            "time_end": 0,
            "train_acc": 0,
            "train_acc_items": [],
            "train_count": 0,
            "train_loss": 0,
            "train_loss_items": [],
            "train_batch_iter": 0,
            "val_acc": 0,
            "val_acc_items": [],
            "val_count": 0,
            "val_loss": 0,
            "val_loss_items": [],
            "val_batch_iter": 0,
            "iter_value": 0,
            "total_count": 0,
            "pos": 0,
            "t": 0,
        }
        
        return status
    
    
    def add_epoch(self, params):
        status = params["status"]
        epoch = status["epoch"]
        self.history[epoch] = status.copy()
    
    
    def on_train_iter(self, params):
        
        status = params["status"]
        train_count = status["train_count"]
        train_acc_items = status["train_acc_items"]
        train_loss_items = status["train_loss_items"]
        
        if self.loss_reduction == "mean":
            if len(train_loss_items) > 0:
                status["train_loss"] = sum(train_loss_items) / len(train_loss_items)
        
        elif self.loss_reduction == "sum":
            if len(train_loss_items) > 0 and train_count > 0:
                status["train_loss"] = sum(train_loss_items) / train_count
        
        if self.acc_reduction == "mean":
            if len(train_acc_items) > 0:
                status["train_acc"] = sum(train_acc_items) / len(train_acc_items)
                status["train_acc_percent"] = status["train_acc"] * 100
        
        elif self.acc_reduction == "sum":
            if len(train_acc_items) > 0 and train_count > 0:
                status["train_acc"] = sum(train_acc_items) / train_count
                status["train_acc_percent"] = status["train_acc"] * 100
        
        if status["total_count"] > 0:
            status["iter_value"] = (status["pos"] / status["total_count"]) * 100
        
    
    def on_val_iter(self, params):
        
        status = params["status"]
        val_count = status["val_count"]
        val_acc_items = status["val_acc_items"]
        val_loss_items = status["val_loss_items"]
        
        if self.loss_reduction == "mean":
            if len(val_loss_items) > 0:
                status["val_loss"] = sum(val_loss_items) / len(val_loss_items)
        
        elif self.loss_reduction == "sum":
            if len(val_loss_items) > 0 and val_count > 0:
                status["val_loss"] = sum(val_loss_items) / val_count
        
        if self.acc_reduction == "mean":
            if len(val_acc_items) > 0:
                status["val_acc"] = sum(val_acc_items) / len(val_acc_items)
                status["val_acc_percent"] = status["val_acc"] * 100
        
        elif self.acc_reduction == "sum":
            if len(val_acc_items) > 0 and val_count > 0:
                status["val_acc"] = sum(val_acc_items) / val_count
                status["val_acc_percent"] = status["val_acc"] * 100
        
        if status["total_count"] > 0:
            status["iter_value"] = (status["pos"] / status["total_count"]) * 100
        
    
    def on_end_epoch(self, params):
        
        self.on_train_iter(params)
        self.on_val_iter(params)
        
        status = params["status"]
        
        lr = []
        for param_group in self.optimizer.param_groups:
            lr.append( param_group['lr'] )
        
        if not(status["train_acc"] is None) and not(status["val_acc"] is None):
            status["rel"] = (status["train_acc"] / status["val_acc"]) if status["val_acc"] > 0 else 0
        
        status["lr"] = lr
        status["lr_str"] = "[" + ",".join([ str(round(item,7)) for item in lr ]) + "]"
    
    
    def get_train_loss(self, epoch=None):
        if epoch is None:
            epoch = self.epoch
        value = self.history[epoch]["train_loss"]
        if value is None:
            value = 0
        return value
    
    
    def get_val_loss(self, epoch=None):
        if epoch is None:
            epoch = self.epoch
        value = self.history[epoch]["train_loss"]
        if value is None:
            value = 0
        return value
    
    
    def get_train_acc(self, epoch=None):
        if epoch is None:
            epoch = self.epoch
        value = self.history[epoch]["train_acc"]
        if value is None:
            value = 0
        return value
    
    
    def get_val_acc(self, epoch=None):
        if epoch is None:
            epoch = self.epoch
        value = self.history[epoch]["train_acc"]
        if value is None:
            value = 0
        return value
    
    
    def get_epoch_metric(self, metric_name, epoch=None):
        if epoch is None:
            epoch = self.epoch
        value = self.history[epoch][metric_name]
        if value is None:
            value = 0
        return value
    
    
    def upload_to_google_drive(self, epoch, repository_path):
        
        import shutil
    
        if not os.path.exists('/content/drive'):
            from google.colab import drive
            drive.mount('/content/drive')
        
        dest_path = os.path.join(repository_path, self.get_model_name())
        if not os.path.exists(dest_path):
            os.makedirs(dest_path)
        
        def upload(file_name):
            src_file_path = os.path.join(self.model_path, file_name)
            if os.path.exists(src_file_path):
                dest_file_path = os.path.join(dest_path, file_name)
                shutil.copy(src_file_path, dest_file_path)

        upload(self.get_model_name() + "-" + str(epoch) + ".data")
        upload(self.get_model_name() + "-" + str(epoch) + ".pth")
    
    
    def download_from_google_drive(self, epoch, repository_path):
        
        import shutil
        
        if not os.path.exists('/content/drive'):
            from google.colab import drive
            drive.mount('/content/drive')
        
        src_path = os.path.join(repository_path, self.get_model_name())
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)
        
        def download(file_name):
            src_file_path = os.path.join(src_path, file_name)
            if os.path.exists(src_file_path):
                dest_file_path = os.path.join(self.model_path, file_name)
                shutil.copy(src_file_path, dest_file_path)

        download(self.get_model_name() + "-" + str(epoch) + ".data")
        download(self.get_model_name() + "-" + str(epoch) + ".pth")
    
    
    def upload_history_to_google_drive(self, repository_path):
    
        import shutil
        
        if not os.path.exists('/content/drive'):
            from google.colab import drive
            drive.mount('/content/drive')
        
        if not os.path.exists(repository_path):
            os.makedirs(repository_path)
        
        dest_path = os.path.join(repository_path, self.get_model_name())
        if not os.path.exists(dest_path):
            os.makedirs(dest_path)
        
        src_file_path = os.path.join(self.model_path, "history.json")
        dest_file_path = os.path.join(dest_path, "history.json")
        shutil.copy(src_file_path, dest_file_path)
    
    
    def download_history_from_google_drive(self, repository_path):
    
        import shutil
        
        if not os.path.exists('/content/drive'):
            from google.colab import drive
            drive.mount('/content/drive')
        
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)
        
        src_path = os.path.join(repository_path, self.get_model_name())
        src_file_path = os.path.join(src_path, "history.json")
        dest_file_path = os.path.join(self.model_path, "history.json")
        shutil.copy(src_file_path, dest_file_path)


class ReAccuracyCallback():
    
    def on_start_epoch(self, params):
        
        self.train_y_batch = []
        self.train_y_pred = []
        self.val_y_batch = []
        self.val_y_pred = []
    
    def on_train_iter(self, params):
        
        self.train_y_batch += params["iter"]["y_batch"].detach().cpu()
        self.train_y_pred += params["iter"]["y_pred"].detach().cpu()

    def on_val_iter(self, params):
        
        self.val_y_batch += params["iter"]["y_batch"].detach().cpu()
        self.val_y_pred += params["iter"]["y_pred"].detach().cpu()
    
    def on_train(self, params):
        
        acc_fn = params["model"].acc_fn
        params["status"]["train_acc_items"] = [
            acc_fn(torch.vstack(self.train_y_pred), torch.vstack(self.train_y_batch))
        ]
        
    def on_val(self, params):
        
        acc_fn = params["model"].acc_fn
        params["status"]["val_acc_items"] = [
            acc_fn(torch.vstack(self.val_y_pred), torch.vstack(self.val_y_batch))
        ]


class SaveCallback():
    
    def __init__(self, count=20, save_weights=True, save_train=True, save_last=False):
        self.count = count
        self.save_weights = save_weights
        self.save_train = save_train
        self.save_last = save_last
    
    def on_save(self, params):
        
        model = params["model"]
        
        is_save = False
        
        if self.save_weights:
            model.save_train_epoch()
            is_save = True
        
        if self.save_train:
            model.save_weights_epoch()
            is_save = True
        
        if self.count >= 0:
            model.save_the_best_models(self.count)
            is_save = True
        
        if self.save_last:
            model.save_model()
            is_save = True
        
        if is_save:
            model.save_history()


class ProgressCallback():
    
    def __init__(self, one_line=False, progress_iter=True, show_lr=True, show_acc=True):
        self.one_line = one_line
        self.progress_iter = progress_iter
        
        self.progress_string_train = [
            "Epoch: {epoch}",
            "{iter_value:.0f}%",
            "train_acc: {train_acc_percent:.2f}%" if show_acc else "",
            "train_loss: {train_loss:.7f}",
            "{t}s",
        ]
        self.progress_string_train = list(filter(lambda item: item != "", self.progress_string_train))
        self.progress_string_train = ", ".join(self.progress_string_train)
        
        self.progress_string_val = [
            "Epoch: {epoch}",
            "{iter_value:.0f}%",
            "val_acc: {val_acc_percent:.2f}%" if show_acc else "",
            "val_loss: {val_loss:.7f}",
            "{t}s",
        ]
        self.progress_string_val = list(filter(lambda item: item != "", self.progress_string_val))
        self.progress_string_val = ", ".join(self.progress_string_val)
        
        self.epoch_string = [
            "Epoch: {epoch}",
            "train_acc: {train_acc_percent:.2f}%" if show_acc else "",
            "val_acc: {val_acc_percent:.2f}%" if show_acc else "",
            "rel: {rel:.3f}" if show_acc else "",
            "train_loss: {train_loss:.7f}",
            "val_loss: {val_loss:.7f}",
            "lr: {lr_str}" if show_lr else "",
            "t: {t}s",
        ]
        self.epoch_string = list(filter(lambda item: item != "", self.epoch_string))
        self.epoch_string = ", ".join(self.epoch_string)
    
    
    def get_status(self, status):
        return status
    
    
    def get_epoch_string(self, status):
        status = self.get_status(status)
        return self.epoch_string.format(**status)
    
    
    def get_progress_string(self, kind, status):
        
        if kind == "train":
            return self.progress_string_train.format(**status)
        
        if kind == "val":
            return self.progress_string_val.format(**status)
    
    
    def on_train_iter(self, params):
        
        status = params["status"]
        
        if self.progress_iter:
            print ("\r" + self.get_progress_string("train", status), end="")
    
    
    def on_val_iter(self, params):
        
        status = params["status"]
        
        if self.progress_iter:
            print ("\r" + self.get_progress_string("val", status), end="")
    
    
    def on_end_epoch(self, params):
        
        status = params["status"]
        
        if self.one_line:
            print( "\r" + self.get_epoch_string(status), end="" )
        else:
            print( "\r" + self.get_epoch_string(status) )
    
    
    def on_end(self, params):
        if self.one_line:
            print ("")
        
        print ("Ok")


class ReloadDatasetCallback():
    
    def on_start_epoch(self, params):
        
        # Train loader
        params["train_loader"] = DataLoader(
            params["train_dataset"],
            batch_size=params["batch_size"],
            collate_fn=params["collate_fn"],
            drop_last=False,
            shuffle=True
        )
        
        # Val loader
        if "val_dataset" in params:
            params["val_loader"] = DataLoader(
                params["val_dataset"],
                batch_size=params["batch_size"],
                collate_fn=params["collate_fn"],
                drop_last=False,
                shuffle=False
            )


class RandomDatasetCallback():

    def __init__(self, train_count, val_count=None):
        self.train_count = train_count
        self.val_count = val_count

    def get_indices(self, total_count, iter_count):
        indices = list(np.random.permutation(total_count))
        return indices[:iter_count]

    def on_start_epoch(self, params):
        
        from torch.utils.data import SubsetRandomSampler
        
        status = params['status']
        batch_size = params['batch_size']

        params['train_loader'] = DataLoader(
            params['train_dataset'],
            batch_size=batch_size,
            collate_fn=params["collate_fn"],
            sampler=SubsetRandomSampler(
                self.get_indices(len(params['train_dataset']), self.train_count)
            )
        )
        
        status["total_count"] = self.train_count
        
        if self.val_count is not None:
            params['val_loader'] = DataLoader(
                params['val_dataset'],
                batch_size=batch_size,
                collate_fn=params["collate_fn"],
                sampler=SubsetRandomSampler(
                    self.get_indices(len(params['val_dataset']), self.val_count)
                )
            )
            
            status["total_count"] += self.val_count




# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##

import torch, math, json, os, re, time
import numpy as np
from torch import nn
from torch.utils.data import Dataset, DataLoader
from PIL import Image, ImageDraw


class TransformDataset(torch.utils.data.Dataset):
    
    def __init__(self, dataset, transform_x=None, transform_y=None):
        self.dataset = dataset
        self.transform_x = transform_x
        self.transform_y = transform_y
        
    def __getitem__(self, index):
        
        x, y = self.dataset[index]
        
        if x is not None and self.transform_x:
            x = self.transform_x(x)
        
        if y is not None and self.transform_y:
            y = self.transform_y(y)
        
        return x, y
        
    def __len__(self):
        return len(self.dataset)


def append_tensor(res, t):
    
    """
    Append tensor
    """
    
    t = t[None, :]
    res = torch.cat( (res, t) )
    return res


def make_index(arr, field_name=None):
    
    """
    Make index from arr. Returns dict of positions values in arr
    """
    
    res = {}
    for index in range(len(arr)):
        value = arr[index]
        if field_name is not None:
            value = value[field_name]
        res[value] = index
    
    return res


def one_hot_encoder(num_class):
    
    """
    Returns one hot encoder to num class
    """
    
    def f(t):
        if not isinstance(t, torch.Tensor):
            t = torch.tensor(t)
        t = nn.functional.one_hot(t.to(torch.int64), num_class).to(torch.float32)
        return t
    
    return f


def label_encoder(labels):
    
    """
    Returns one hot encoder from label
    """
    
    labels = make_index(labels)
    
    def f(label_name):
        
        index = labels[label_name] if label_name in labels else -1
        
        if index == -1:
            return torch.zeros( len(labels) )
        
        t = torch.tensor(index)
        return nn.functional.one_hot(t.to(torch.int64), len(labels)).to(torch.float32)
    
    return f


def bag_of_words_encoder(dictionary_sz):
    
    """
    Returns bag of words encoder from dictionary indexes.
    """
    
    def f(text_index):
        
        t = torch.zeros(dictionary_sz - 1)
        for index in text_index:
            if index > 0:
                t[index - 1] = 1
        
        return t
        
    return f


def dictionary_encoder(dictionary, max_words):
    
    """
    Returns one hot encoder from text.
    In dictionary 0 pos is empty value, if does not exists in dictionary
    """
    
    def f(text_arr):
        
        t = torch.zeros(max_words).to(torch.int64)
        text_arr_sz = min(len(text_arr), max_words)
        
        pos = 0
        for i in range(text_arr_sz):
            word = text_arr[i]
            
            if word in dictionary:
                index = dictionary[word]
                t[pos] = index
                pos = pos + 1
        
        return t
    
    return f


def batch_map(f):
    
    def transform(batch_x):
        
        res = torch.tensor([])
        
        for i in range(len(batch_x)):
            x = f(batch_x[i])
            x = x[None, :]
            res = torch.cat( (res, x) )
        
        return res.to(batch_x.device)
    
    return transform


def batch_to(x, device):
    
    """
    Move batch to device
    """
    
    if isinstance(x, list) or isinstance(x, tuple):
        for i in range(len(x)):
            x[i] = x[i].to(device)
    else:
        x = x.to(device)
    
    return x


def tensor_size(t):

    """
    Returns tensor size
    """
    
    if not isinstance(t, torch.Tensor):
        return 0, 0
    
    sz = t.element_size()
    shape = t.shape
    params = 1

    for c in shape:
        params = params * c

    size = params * sz

    return params, size


def create_dataset_indexes(dataset, file_name):
    
    """
    Load dataset indexes
    """
    
    index = load_json( file_name )
    if index is None:
        index = list(np.random.permutation( len(dataset) ))
        save_json(file_name, index, indent=None)
    
    return index


def split_dataset(dataset, k=0.2, indexes=None):
    
    """
    Split dataset for train and validation
    """
    
    sz = len(dataset)
    train_count = round(sz * (1 - k))
    val_count = sz - train_count
    
    if indexes is None:
        indexes = list(np.random.permutation(sz))
    
    from torch.utils.data import Subset
    return [Subset(dataset, indexes[0 : train_count]), Subset(dataset, indexes[train_count : ])]
    


def get_default_device():
    
    """
    Returns default device
    """
    
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    
    return device


def get_acc_class():
    
    def f(batch_predict, batch_y):
        
        """
        Returns class accuracy
        """
        
        if len(batch_y.shape) == 2:
            batch_y = torch.argmax(batch_y, dim=1)
        
        batch_predict = torch.argmax(batch_predict, dim=1)
        acc = torch.sum( torch.eq(batch_y, batch_predict) ).item()
        
        return acc
    
    return f


def get_acc_binary(treshold=0.5):
    
    def f(batch_predict, batch_y):
        
        batch_predict = (batch_predict >= treshold) * 1.0
        acc = torch.sum( torch.eq(batch_y, batch_predict) ).item()
        
        if len(batch_y.shape) == 2:
            return acc / batch_y.shape[1]
            
        return acc
        
    return f
    
    
    """
    Returns binary accuracy
    """
    
    from torcheval.metrics import BinaryAccuracy
    
    batch_predict = batch_predict.reshape(batch_predict.shape[0])
    batch_y = batch_y.reshape(batch_y.shape[0])
    
    acc = BinaryAccuracy() \
        .to(batch_predict.device) \
        .update(batch_predict, batch_y) \
        .compute().item()
    
    return round(acc * len(batch_y))


def resize_image(image, new_size, contain=True, color=None):
   
    """
    Resize image
    """
    
    if isinstance(image, str):
        image = Image.open(image)
    
    if isinstance(image, np.ndarray):
        image = Image.fromarray(image)
    
    w1 = image.size[0]
    h1 = image.size[1]
    w2 = new_size[0]
    h2 = new_size[1]

    k1 = w1 / h1
    k2 = w2 / h2
    w_new = 0
    h_new = 0
    
    if k1 > k2 and contain or k1 < k2 and not contain:
        h_new = round(w2 * h1 / w1)
        w_new = w2
        
    else:
        h_new = h2
        w_new = round(h2 * w1 / h1)
    
    image_new = image.resize( (w_new, h_new) )
    image_resize = resize_image_canvas(image_new, new_size, color=color)
    del image_new
    
    return image_resize
    

def resize_image_canvas(image, size, color=None):
   
    """
    Resize image canvas
    """
    
    width, height = size
    
    if color == None:
        pixels = image.load()
        color = pixels[0, 0]
        del pixels
    
    image_new = Image.new(image.mode, (width, height), color = color)
    
    position = (
        math.ceil((width - image.size[0]) / 2),
        math.ceil((height - image.size[1]) / 2),
    )
    
    image_new.paste(image, position)
    return image_new


def load_image(file_name, convert=None, load_as=""):
    
    image = Image.open(file_name)
    
    if convert:
        image = image.convert(convert)
    
    if load_as == "numpy":
        image = np.array(image)
    
    elif load_as == "torch":
        image = torch.from_numpy(np.array(image))
    
    return image


def draw_images(images, ncols=4, cmap=None, first_channel=False):

    import matplotlib.pyplot as plt
    
    count_images = len(images)
    
    fig, ax = plt.subplots(
        (count_images - 1) // ncols + 1, ncols,
        figsize=(10, 6), sharey=True, sharex=True
    )
    fig.subplots_adjust(0, 0, 1, 1)
    ax = ax.flatten()

    # Рисуем картинки
    for i in range(count_images):
        image = images[i]
        
        if isinstance(image, str):
            image = load_image(image, load_as="torch")
        
        if isinstance(image, np.ndarray):
            pass
        
        if isinstance(image, torch.Tensor):
            if first_channel == True:
                image = move_rgb_to_end(image)
        
        ax[i].imshow(image, cmap)

    # Скрываем лишние картинки
    for i in range(count_images, len(ax)):
        ax[i].set_visible(False)
    
    plt.show()


def draw_images_grid(images, ncols=4, first_channel=False, *args, **kwargs):
    
    import matplotlib.pyplot as plt
    from torchvision.utils import make_grid
    
    if isinstance(images, list):
        images = images.copy()
        for index in range(len(images)):
            
            if isinstance(images[index], Image.Image):
                images[index] = torch.from_numpy( np.array(images[index]) )
            
            if isinstance(images[index], np.ndarray):
                images[index] = torch.from_numpy( images[index] )
            
            if not first_channel:
                images[index] = move_rgb_to_begin(images[index])
    
    elif isinstance(images, torch.Tensor):
        if not first_channel:
            images = move_rgb_to_begin(images)
    
    plt.imshow(move_rgb_to_end(make_grid(images, nrow=ncols, *args, **kwargs)))
    plt.show()


def draw_image(image, cmap=None, first_channel=False):
    
    """
    Plot show image
    """
    
    if isinstance(image, str):
        image = Image.open(image)
    
    elif isinstance(image, np.ndarray):
        pass
    
    elif torch.is_tensor(image):
        if first_channel == True:
            image = move_rgb_to_end(image)
    
    import matplotlib.pyplot as plt
    
    plt.imshow(image, cmap)
    plt.show()


def move_rgb_to_end(t):
    l = len(t.shape)
    t = torch.moveaxis(t, l-3, l-1)
    return t


def move_rgb_to_begin(t):
    l = len(t.shape)
    t = torch.moveaxis(t, l-1, l-3)
    return t

def swap_hw(t):
    l = len(t.shape)
    t = torch.moveaxis(t, l-1, l-2)
    return t


def list_files(path="", recursive=True, full_path=False):
    
    """
        Returns files in folder
    """
    
    def read_dir(path, recursive=True):
        res = []
        items = os.listdir(path)
        for item in items:
            
            item_path = os.path.join(path, item)
            
            if item_path == "." or item_path == "..":
                continue
            
            if os.path.isdir(item_path):
                if recursive:
                    res = res + read_dir(item_path, recursive)
            else:
                res.append(item_path)
            
        return res
    
    try:
        items = read_dir( path, recursive )
            
        def f(item):
            return item[len(path + "/"):]
        
        items = list( map(f, items) )
    
    except Exception:
        items = []
    
    if full_path:
        items = list( map(lambda x: os.path.join(path, x), items) )
    
    return items


def get_sort_alphanum_key(name):
    
    """
    Returns sort alphanum key
    """
    
    arr = re.split("([0-9]+)", name)
    
    for key, value in enumerate(arr):
        try:
            value = int(value)
        except:
            pass
        arr[key] = value
    
    arr = list(filter(lambda item: item != "", arr))
    
    return arr


def alphanum_sort(files):
    
    """
    Alphanum sort
    """
    
    files.sort(key=get_sort_alphanum_key)


def list_dirs(path=""):
    
    """
        Returns dirs in folder
    """
    
    try:
        items = os.listdir(path)
    except Exception:
        items = []
    
    return items


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def save_json(file_name, obj, indent=2):
    
    """
    Save json to file
    """
    
    json_str = json.dumps(obj, indent=indent, cls=JSONEncoder, ensure_ascii=False)
    file = open(file_name, "w")
    file.write(json_str)
    file.close()


def load_json(file_name):
    
    """
    Load json from file
    """
    
    obj = None
    file = None
    
    try:
        
        file = open(file_name, "r")
        s = file.read()
        obj = json.loads(s)
        
    except Exception:
        pass
    
    finally:
        if file:
            file.close()
            file = None
    
    return obj


def load_epoch(model, model_name, epoch, repository_path="model"):
        
    """
    Load epoch
    """
    
    model_path = os.path.join(repository_path, model_name)
    
    file_name = model_name + "-" + str(epoch) + ".data"
    file_path = os.path.join(model_path, file_name)
    
    if not os.path.exists(file_path):
        file_name = model_name + "-" + str(epoch) + ".pth"
        file_path = os.path.join(model_path, file_name)
    
    load_model_from_file(model, file_path)


def load_model(model, model_name, file_name, repository_path="model"):
    
    """
    Load model
    """
    
    model_path = os.path.join(repository_path, model_name)
    file_path = os.path.join(model_path, file_name)
    
    if not os.path.exists(file_path):
        file_path = os.path.join(model_path, file_name)
    
    load_model_from_file(model, file_path)
    

def load_model_from_file(model, file_path):
        
    """
    Load model from file
    """
    
    save_metrics = torch.load(file_path)
    state_dict = save_metrics
    
    if "epoch" in save_metrics:
        if isinstance(model, nn.Module):
            state_dict = save_metrics["module"]
    
    model.load_state_dict(state_dict, strict=False)


def summary(module, x, model_name=None, device=None, batch_size=2, collate_fn=None):
        
    """
    Show model summary
    """
    
    hooks = []
    layers = []
    res = {
        "params_count": 0,
        "params_train_count": 0,
        "total_size": 0,
    }
    
    def forward_hook(module, input, output):
        
        output = output[0] if isinstance(output, tuple) else output
        output_shape = "(?)"
        if hasattr(output, "shape") and isinstance(output, torch.Tensor):
            output_shape = output.shape
            
        class_name = module.__class__.__module__ + "." + module.__class__.__name__
        layer = {
            "name": module.__class__.__name__,
            "class_name": module.__class__.__module__ + "." + module.__class__.__name__,
            "shape": output_shape,
            "params": 0
        }
        
        if layer["name"] == model_name:
            layer["name"] = "Output"
        
        # Calc parameters
        for name, p in module.named_parameters():
            
            if "." in name:
                continue
            
            params = p.numel()
            size = p.numel() * p.element_size()
            
            res["params_count"] += params
            res["total_size"] += size
            layer["params"] += params
            
            if p.requires_grad:
                res["params_train_count"] += params
        
        # Add output size
        params, size = tensor_size(output)
        #res["total_size"] += size
        
        # Add layer
        layers.append(layer)
            
    def add_hooks(module):
        hooks.append(module.register_forward_hook(forward_hook))
    
    # Get batch from Dataset
    batch = None
    if isinstance(x, torch.utils.data.Dataset):
        loader = torch.utils.data.DataLoader(
            x,
            batch_size=batch_size,
            collate_fn=collate_fn,
            drop_last=False,
            shuffle=False
        )
        it = loader._get_iterator()
        
        batch = next(it)
        
        if hasattr(module, "batch_transform"):
            batch = module.batch_transform(batch, device)
        
        x = batch["x"]
    
    # Add input size
    if isinstance(x, list) or isinstance(x, tuple):
        shapes = []
        for i in range(len(x)):
            params, size = tensor_size(x[i])
            shapes.append(x[i].shape)
        #res["total_size"] += size
        layers.append({
            "name": "Input",
            "shape": shapes,
            "params": 0
        })
    
    else:
        params, size = tensor_size(x)
        #res["total_size"] += size
        layers.append({
            "name": "Input",
            "shape": x.shape,
            "params": 0
        })
    
    module.apply(add_hooks)
    
    if hasattr(module, "step_forward"):
        _, _, y = module.step_forward(batch, device)
    
    else:
        
        # Move to device
        if device is not None:
            x = batch_to(x, device)
        
        # Module predict
        with torch.no_grad():
            module.eval()
            y = module(x)
    
    # Clear cache
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    
    # Remove hooks
    for item in hooks:
        item.remove()
    
    res['total_size'] = round(res['total_size'] / 1024 / 1024 * 100) / 100
    
    # Calc info
    values = []
    for i, layer in enumerate(layers):
        shape = layer["shape"]
        shape_str = ""
        if isinstance(shape, list):
            shape = [ "(" + ", ".join(map(str,s)) + ")" for s in shape ]
            shape_str = "[" + ", ".join(shape) + "]"
        else:
            shape_str = "(" + ", ".join(map(str,shape)) + ")"
        
        values.append([i + 1, layer["name"], shape_str, layer["params"]])
    
    # Print info
    info_sizes = [2, 7, 7, 5]
    for _, value in enumerate(values):
        for i in range(4):
            sz = len(str(value[i]))
            if info_sizes[i] < sz:
                info_sizes[i] = sz
        
    def format_row(arr, size):
        s = "{:<"+str(size[0] + 1)+"} {:>"+str(size[1] + 2)+"}" + \
            "{:>"+str(size[2] + 5)+"} {:>"+str(size[3] + 5)+"}"
        return s.format(*arr)
    
    width = info_sizes[0] + 1 + info_sizes[1] + 2 + info_sizes[2] + 5 + info_sizes[3] + 5 + 2
    print( "=" * width )
    print( format_row(["", "Layer", "Output", "Params"], info_sizes) )
    print( "-" * width )
    
    for _, value in enumerate(values):
        print( format_row(value, info_sizes) )
    
    print( "-" * width )
    if model_name is not None:
        print( f"Model name: {model_name}" )
    print( f"Total params: {res['params_count']:_}".replace('_', ' ') )
    if res['params_count'] != res['params_train_count'] and res['params_train_count'] > 0:
        print( f"Trainable params: {res['params_train_count']:_}".replace('_', ' ') )
    print( f"Total size: {res['total_size']} MiB" )
    print( "=" * width )


def compile(module):
    from .Model import Model
    return Model(module)


def fit(
    model, train_dataset=None, val_dataset=None,
    batch_size=64, epochs=10, collate_fn=None,
    callbacks=None, do_train=True, do_val=True,
    **params
):
    if callbacks is None:
        callbacks = []
    
    callbacks.insert(0, model.module)
    callbacks.insert(0, model)
    
    params["model"] = model
    params["train_dataset"] = train_dataset
    params["val_dataset"] = val_dataset
    params["batch_size"] = batch_size
    params["epochs"] = epochs
    params["status"] = None
    params["callbacks"] = callbacks
    params["collate_fn"] = collate_fn
    params["iter"] = {}
    
    device = model.device
    model_name = model.get_model_name()
    
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    
    # Train loader
    if not "train_loader" in params:
        train_loader = DataLoader(
            train_dataset,
            batch_size=batch_size,
            collate_fn=collate_fn,
            drop_last=False,
            shuffle=True
        )
        params["train_loader"] = train_loader
    
    # Val loader
    val_loader = None
    if val_dataset:
        if not "val_loader" in params:
            val_loader = DataLoader(
                val_dataset,
                batch_size=batch_size,
                collate_fn=collate_fn,
                drop_last=False,
                shuffle=False
            )
            params["val_loader"] = val_loader
    
    acc_fn = model.acc_fn
    device = model.device
    loss_fn = model.loss
    min_lr = model.min_lr
    module = model.module
    optimizer = model.optimizer
    scheduler = model.scheduler
    
    batch_transform = getattr(module, "batch_transform", None)
    step_forward = getattr(module, "step_forward", None)
    step_loss = getattr(module, "step_loss", None)
    step_scheduler = getattr(module, "step_scheduler", None)
    get_batch_size = getattr(module, "get_batch_size", None)
    
    if isinstance(loss_fn, nn.Module):
        loss_fn = loss_fn.to(model.device)
    
    def call_callback(name, params):
        if callbacks is not None:
            for callback in callbacks:
                if hasattr(callback, name):
                    f = getattr(callback, name, None)
                    if f is not None:
                        f(params)
    
    
    call_callback("on_start", params)
    
    print ("Start train " + str(model_name) + " on " + str(device))
    try:
        while model.do_training(epochs):
            
            time_start = time.time()
            
            params["status"] = model.get_epoch_train_status()
            
            params["status"]["total_count"] = len(train_dataset)
            if val_dataset is not None:
                params["status"]["total_count"] += len(val_dataset)
            
            call_callback("on_start_epoch", params)
            
            train_loader = params["train_loader"]
            val_loader = params["val_loader"]
            
            if do_train != False:
                
                # Train mode
                model.train()
                
                for batch in train_loader:
                    
                    batch_len = 1
                    if get_batch_size is not None:
                        batch_len = get_batch_size(batch)
                    else:
                        batch_len = len(batch["x"])
                    
                    # Set parameter gradients to zero
                    optimizer.zero_grad()
                    
                    acc_value = None
                    loss = None
                    
                    # Forward train
                    if step_forward is not None:
                        loss, acc_value, _ = step_forward(
                            batch, device,
                            loss_fn=loss_fn, acc_fn=acc_fn
                        )
                    
                    else:
                        
                        # Data to device
                        if batch_transform:
                            batch = batch_transform(batch, device)
                        
                        x_batch = batch_to(batch["x"], device)
                        y_batch = batch_to(batch["y"], device)
                        y_pred = module(x_batch)
                        
                        # Calc loss
                        if step_loss is not None:
                            loss = step_loss(y_pred, y_batch, loss_fn=loss_fn)
                        else:
                            loss = loss_fn(y_pred, y_batch, loss_fn=loss_fn)
                        
                        # Calc accuracy
                        if acc_fn is not None:
                            acc_value = acc_fn(y_pred, y_batch)
                        
                        params["iter"]["x_batch"] = x_batch
                        params["iter"]["y_batch"] = y_batch
                        params["iter"]["y_pred"] = y_pred
                        
                        del x_batch, y_batch, y_pred
                    
                    # Backward
                    loss.backward()
                    optimizer.step()
                    
                    # Add status
                    params["status"]["pos"] += batch_len
                    params["status"]["train_count"] += batch_len
                    params["status"]["train_loss_items"].append( loss.item() )
                    params["status"]["t"] = round(time.time() - time_start)
                    
                    if acc_value is not None:
                        params["status"]["train_acc_items"].append( acc_value )
                    
                    call_callback("on_train_iter", params)
                    
                    # Clear cache
                    if "x_batch" in params["iter"]:
                        del params["iter"]["x_batch"]
                    
                    if "y_batch" in params["iter"]:
                        del params["iter"]["y_batch"]
                    
                    if "y_pred" in params["iter"]:
                        del params["iter"]["y_pred"]
                    
                    del loss
                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
            
            call_callback("on_train", params)
            
            if val_loader is not None and do_val != False:
                
                # Testing mode
                model.eval()
                
                with torch.no_grad():

                    for batch in val_loader:
                        
                        batch_len = 1
                        if get_batch_size is not None:
                            batch_len = get_batch_size(batch)
                        else:
                            batch_len = len(batch["x"])
                        
                        acc_value = None
                        loss = None
                        
                        # Forward
                        if step_forward is not None:
                            loss, acc_value, _ = step_forward(
                                batch, device,
                                loss_fn=loss_fn, acc_fn=acc_fn
                            )
                            
                        else:
                            
                            # data to device
                            if batch_transform:
                                batch = batch_transform(batch, device)
                            
                            x_batch = batch_to(batch["x"], device)
                            y_batch = batch_to(batch["y"], device)
                            y_pred = module(x_batch)
                            
                            # Calc loss
                            if step_loss is not None:
                                loss = step_loss(y_pred, y_batch)
                            else:
                                loss = loss_fn(y_pred, y_batch)
                            
                            # Calc accuracy
                            if acc_fn is not None:
                                acc_value = acc_fn(y_pred, y_batch)
                            
                            params["iter"]["x_batch"] = x_batch
                            params["iter"]["y_batch"] = y_batch
                            params["iter"]["y_pred"] = y_pred
                            
                            del x_batch, y_batch, y_pred
                        
                        # Add status
                        params["status"]["pos"] += batch_len
                        params["status"]["val_count"] += batch_len
                        params["status"]["val_loss_items"].append( loss.item() )
                        params["status"]["t"] = round(time.time() - time_start)
                        
                        if acc_value is not None:
                            params["status"]["val_acc_items"].append( acc_value )
                        
                        call_callback("on_val_iter", params)
                        
                        # Clear cache
                        if "x_batch" in params["iter"]:
                            del params["iter"]["x_batch"]
                        
                        if "y_batch" in params["iter"]:
                            del params["iter"]["y_batch"]
                        
                        if "y_pred" in params["iter"]:
                            del params["iter"]["y_pred"]
                        
                        del loss
                        if torch.cuda.is_available():
                            torch.cuda.empty_cache()
            
            call_callback("on_val", params)
            call_callback("on_next_epoch", params)
            
            # Add metricks
            time_end = time.time()
            params["status"]["t"] = round(time_end - time_start)
            params["status"]["time_end"] = time_end
            
            call_callback("on_end_epoch", params)
            
            if scheduler is not None:
                if step_scheduler is not None:
                    step_scheduler(params)
                else:
                    scheduler.step()
            
            model.add_epoch(params)
            call_callback("on_save", params)
            
            model.epoch = model.epoch + 1
            
        call_callback("on_end", params)
        
    except KeyboardInterrupt:
        
        print ("")
        print ("Stopped manually")
        print ("")

    if torch.cuda.is_available():
        torch.cuda.empty_cache()


def save_embeddings(dataset, file_name, transform, emb_size, batch_size=8):
    
    import h5py, gc
    
    loader = DataLoader(
        dataset,
        batch_size=batch_size,
        drop_last=False,
        shuffle=False
    )
    
    pos = 0
    next_pos = 0
    dataset_count = len(dataset)
    time_start = time.time()
    
    if os.path.exists(file_name):
        os.remove(file_name)
    
    with h5py.File(file_name, 'w') as file:
        
        file_dataset = file.create_dataset(
            'data', (dataset_count, emb_size), dtype='float32'
        )
        
        for batch in loader:
            
            batch = transform(batch)
            file_dataset[pos:pos+len(batch["x"])] = batch["x"].cpu().numpy()
            
            # Show progress
            pos = pos + len(batch["x"])
            if pos > next_pos:
                next_pos = pos + 16
                t = str(round(time.time() - time_start))
                print ("\r" + str(math.floor(pos / dataset_count * 100)) + "% " + t + "s", end='')
            
            del batch
            
            # Clear cache
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
            
            gc.collect()
            file.flush()
        
        file.close()


def colab_upload_file_to_google_drive(src, dest):
    
    import shutil
    
    if not os.path.exists('/content/drive'):
        from google.colab import drive
        drive.mount('/content/drive')
    
    dest_dir_path = os.path.dirname(dest)
    if not os.path.exists(dest_dir_path):
        os.makedirs(dest_dir_path)
    
    shutil.copy(src, dest)
    

class ListDataset(Dataset):
    
    def __init__(self, items):
        self.items = items
    
    def __getitem__(self, index: int):
        return self.items[index]
    
    def __len__(self) -> int:
        return len(self.items)



# -*- coding: utf-8 -*-

##
# Tiny ai helper
# Copyright (с) Ildar Bikmamatov 2022 - 2023 <support@bayrell.org>
# License: MIT
##


class CSVReader:
    
    def __init__(self, file_name, encoding):
        self.file = open(file_name, 'rb')
        self.encoding = encoding
        self.lines = []
        self.file.seek(0, 2)
        self.file_size = self.file.tell()
        self.file.seek(0, 0)
        self.read_header()
        self.read_file()
        
    def read_header(self):
        line = self.file.readline().decode(self.encoding)
        self.header = line.strip().split(",")
        self.header = [ s.strip() for s in self.header ]
    
    def read_file(self):
        start = self.file.tell()
        while start < self.file_size:
            line = self.file.readline()
            end = self.file.tell()
            self.lines.append( (start, end - start) )
            start = end
    
    def __del__(self):
        del(self.lines)
    
    def __len__(self):
        return len(self.lines)
    
    def __getitem__(self, index):
        start, length = self.lines[index]
        self.file.seek(start, 0)
        line = self.file.read(length)
        line = line.decode(self.encoding).strip().split(",")
        line = [ s.strip() for s in line ]
        
        obj = {}
        for i in range(len(line)):
            key = self.header[i]
            value = line[i]
            obj[key] = value
        
        return obj
